#ifndef EBSMAGNET_MAGNET_H
#define EBSMAGNET_MAGNET_H
#include <Quadrupole.h>
#include "EbsMagnet.h"

namespace EbsMagnet_ns {

class EbsMagnet;
class DynAttribute;

typedef struct ConfigItem {
  string name;
  string value;
} ConfigItem;

union value {
  Tango::DevBoolean bValue;
  Tango::DevDouble dValue;
};

typedef struct AttributeItem {

  DynAttribute *handle;  // Use only at init time
  string name;
  value read;
  value set;

} AttributeItem;

class Magnet {

public:

  // Create a Magnet
  static Magnet *Create(EbsMagnet *ds, bool focusing,double crossTalk, string &serial, string &type, string &model);

  // Constructor
  Magnet(EbsMagnet *ds, string& type, string& serial, string& modelName);
  ~Magnet();

  // Return attributes of this magnet
  // Use only at initialisation time, Tango may delete them
  vector<DynAttribute *> getAttributes();

  // Return setpointCheckCommand
  vector<string> getSetpointCheckCommands();

  // Add attribute to the magnet
  AttributeItem *addAttribute(const string& name,long data_type,Tango::AttrWriteType rwType,const string& unit,
                              const string& format,bool memorized,bool memorized_init);

  // Has a main current ?
  bool hasMainCurrent() {
    return model->has_main_current();
  }

  // Number of strengths
  size_t getStrengthNumber() {
    return nbStrengths;
  }

  // Number of currents
  size_t getCurrentNumber() {
    return nbCurrents;
  }

  // Read attribute
  void readAttribute(Tango::Attribute &attr,string& name,void *readValue=nullptr);
  void readStrengths(double *strengths);

  // Write attribute
  void writeAttribute(double setValue,const string& name,vector<string>& psNames,vector<Tango::DevState>& psStates);
  void writeAttribute(bool setValue,const string& name,vector<string>& psNames,vector<Tango::DevState>& psStates);
  void writeStrengths(vector<double>& strengths);

  // Main is frozen
  bool isMainFrozen();

  // Setpoint check
  bool setpointCheck(double strength,string& name);
  bool setpointCheck(vector<double>& strengths);

  // Set currents (update
  void setCurrentsAndUpdateStrength(vector<double>& currents, vector<double>& setCurrents);

  // Compute pseudo current (SH only)
  void computePseudoCurrentsFromCurrents(vector<double>& currents, vector<double>& pseudoCurrents);
  void computeStrengthFromPseudoCurrent(int idx,double& pseudoCurrent, double& strength);

  // Set design strength
  bool setDesignStrength(double strength);

  // Reset resonance strength
  void resetResonanceStrength();

  // Get correction strength setpoint
  double getCorrectionStrength();

  // Get resonance strength setpoint
  double getResonanceStrength();
  void setResonanceStrength(double set);

  // send currents to PS
  void send_currents();

  // Util split function
  static void split(std::vector<std::string> &tokens, const std::string &text, char sep);
  static bool compare(const string& s1,const string& s2);
  static string getStrengthPrefix(const string& name);

  
protected:

  EbsMagnet *ds;
  string modelName;
  string serial;
  size_t nbStrengths;
  size_t nbCurrents;
  string type;

  // shortcuts
  AttributeItem *currentAttr;
  AttributeItem *strengthAtt;
  AttributeItem *designStrengthsAtt;
  AttributeItem *correctionStrengthAtt;
  AttributeItem *resonanceStrengthAtt;
  AttributeItem *frozenAtt;

  // Strength/Current calculation
  MagnetModel::Magnet *model;   // Model
  vector<ConfigItem> config;    // Model data
  void parseConfig();
  string getConfigItem(const string& name);
  double getConfigItemDouble(const string& name);

  // Attributes
  void createAttributes();
  vector<AttributeItem> scalarAttributes;
  AttributeItem *getAttribute(const string& name);

  // Update correction strength
  void updateCorrectionStrength();

  // Check PS setpoints
  bool check_minmax(vector<double>& currents,bool throwException = false);

  // Freeze check
  void frozenCheck(const string& name);

  // Check underlying PS
  void psCheck(const string& name,vector<string>& names,vector<Tango::DevState>& states);

  // Utils functions
  static bool isStrengthH(const string& name);
  static bool isStrengthV(const string& name);
  static bool isStrengthSQ(const string& name);
  static bool isStrength(const string& name);
  static bool isStrengths(const string& name);
  static bool isCurrent(const string& name);
  static bool isMainCurrent(const string& name);
  static bool isFrozen(const string& name);
  static bool isResonanceStrength(const string& name);
  static bool isCorrectionStrength(const string& name);

};

} // end namespace EbsMagnet_ns

#endif //EBSMAGNET_MAGNET_H
